import pytest
import vestaboard


def test_enabling_local_api_succeeds(patched_post):
    b = vestaboard.Board(localApi={"enablementToken": "fake", "ip": "10.0.0.1"})
    assert b.localOptions["useSavedToken"] == True


def test_using_manual_key_and_ip():
    b = vestaboard.Board(localApi={"key": "fakeKey", "ip": "10.0.0.1"})
    assert b.localKey == "fakeKey"
    assert b.localIP == "10.0.0.1"


def test_using_saved_key_and_ip(with_token_file):
    b = vestaboard.Board(localApi={"useSavedToken": True})
    assert b.localKey == "fakeLocalKey"
    assert b.localIP == "10.0.0.2"


def test_warns_when_both_key_and_enablement_token(patched_post):
    with pytest.warns(
        UserWarning,
        match="An enablement token was provided, so I'm enabling the local API for you. If you've already enabled the local API on your board, remove the enablement token and try again.",
    ):
        b = vestaboard.Board(
            localApi={"enablementToken": "token", "ip": "10.0.0.1", "key": "fakeKey"}
        )
        assert b.localIP == "10.0.0.1", "Local IP should be the passed-in IP"
        assert (
            b.localKey == "fakelocalkey"
        ), "Should use the new key the board generated"


def test_errors_when_no_ip_provided_to_enable_api():
    with pytest.raises(ValueError):
        vestaboard.Board(localApi={"enablementToken": "token"})


def test_errors_when_no_ip_provided_to_use_api():
    with pytest.raises(ValueError):
        vestaboard.Board(localApi={"key": "token"})


def test_errors_when_no_saved_key_and_none_provided(no_token_file):
    with pytest.raises(ValueError):
        vestaboard.Board(localApi={})


def test_errors_with_save_only(no_token_file):
    with pytest.raises(ValueError):
        vestaboard.Board(localApi={"saveToken": True})


def test_nonexistent_token_file(no_token_file):
    with pytest.raises(FileNotFoundError):
        vestaboard.Board(localApi={"useSavedToken": True})


def test_fails_with_ip_only(no_token_file):
    with pytest.raises(ValueError):
        vestaboard.Board(localApi={"ip": "10.0.0.1"})


def test_should_allow_reading(with_token_file, patched_get):
    b = vestaboard.Board(localApi={"useSavedToken": True})
    chars = b.read()
    assert chars == {
        "message": [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    }


def test_converts_back_into_characters(with_token_file, patched_get):
    b = vestaboard.Board(localApi={"useSavedToken": True})
    chars = b.read({"convert": True})
    # fmt: off
    assert chars == [
                    ['a', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                    ['b', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                    ['c', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                    ['d', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                    ['e', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                    ['f', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '],
                ]
    # fmt: on


def test_joins_back_into_lines_with_normalize(with_token_file, patched_get):
    b = vestaboard.Board(localApi={"useSavedToken": True})
    chars = b.read({"convert": True, "normalize": True})
    assert (
        chars
        == "a                     \n"
        + "b                     \n"
        + "c                     \n"
        + "d                     \n"
        + "e                     \n"
        + "f                     "
    )

def test_uses_provided_ip_address_when_using_saved_token(with_token_file, patched_get):
    b = vestaboard.Board(localApi={"useSavedToken": True, "ip": "10.10.10.10"})
    assert b.localIP == "10.10.10.10"





subscription = "https://platform.vestaboard.com/subscriptions"
post = "https://platform.vestaboard.com/subscriptions/{}/message"
enableLocal = "http://{}:7000/local-api/enablement"
postLocal = "http://{}:7000/local-api/message"
readWrite = "https://rw.vestaboard.com"



import warnings
import textwrap
import re
import math
from vestaboard.characters import characters
from vestaboard.characters import colors
from vestaboard.characters import reverseCharacters


class Formatter:
    def __init__(self):
        self.name = "Formatter"

    def _standard(self, text):
        if not self._isValid(text):
            raise ValueError(
                "Your text contains one or more characters that the Vestaboard does not support."
            )
        return {"text": text}

    @staticmethod
    def _raw(charList):
        return {"characters": charList}

    @staticmethod
    def _isValid(inputString):
        inputString = inputString.lower()
        test = r"^(?:[A-Za-z\d!@#$()\-+&=;:'\"%,./?°\s]|(?:\{\d{1,2}\}))*$"

        return bool(re.match(test, inputString))

    @staticmethod
    def _getEmbeddedCharCodes(inputString):
        test = r"\{[0-9]+\}+"

        return re.findall(test, inputString)

    def _numCharacterCodes(self, inputString):
        embeddedCharacterCodes = self._getEmbeddedCharCodes(inputString)
        numCharacterCodes = 0
        for match in embeddedCharacterCodes:
            numCharacterCodes += 2 + (len(match) - 3)

        return numCharacterCodes

    def convert(self, inputString, byLetter=True, byWord=False):
        if not self._isValid(inputString):
            raise Exception(
                "Your text contains one or more characters that the Vestaboard does not support."
            )
        inputString = inputString.lower()
        converted = []
        if byWord:
            wordList = inputString.split(" ")
            for word in wordList:
                convertedWord = self._convertLoop(word)
                converted.append(convertedWord)
        elif byLetter:
            converted += self._convertLoop(inputString)

        return converted

    def convertLine(self, inputString, justify="center", color=" ", spaceBuffer=False):
        if not self._isValid(inputString):
            raise Exception(
                "Your text contains one or more characters that the Vestaboard does not support."
            )
        numCharacterCodes = self._numCharacterCodes(inputString)
        if spaceBuffer:
            inputString = self._addSpaceBuffer(inputString, justify)
        inputString = inputString.lower()
        if color != " ":
            try:
                color = colors[color]
            except KeyError as exc:
                raise KeyError(
                    "Valid colors are red, orange, yellow, green, blue, violet, white, and black (default black)."
                ) from exc
        converted = []
        if len(inputString) - numCharacterCodes > 22:
            raise Exception(
                f"Convert line method takes in a string less than or equal to 22 characters - string passed in was {len(inputString)} characters. Reduce size and try again (remember that setting spaceBuffer=True increases your line size by 2)."
            )
        inputString = self._justifyContent(
            inputString, justify, numCharacterCodes, color
        )

        converted = self._convertLoop(inputString)

        return converted

    @staticmethod
    def _convertLoop(inputString):
        converted = []
        skipTo = 0
        for index, letter in enumerate(inputString):
            if index < skipTo:
                continue
            if letter == "{":
                if inputString[index + 2] == "}":  # one-digit character code like {4}
                    converted.append(int(inputString[index + 1 : index + 2]))
                    skipTo = index + 3
                elif (
                    inputString[index + 3] == "}"
                ):  # two-digit character code like {63}
                    converted.append(int(inputString[index + 1 : index + 3]))
                    skipTo = index + 4
            else:
                converted.append(characters[letter])

        return converted

    @staticmethod
    def _addSpaceBuffer(inputString, justify):
        if justify == "left":
            return inputString + " "
        elif justify == "right":
            return " " + inputString
        else:
            return " " + inputString + " "

    @staticmethod
    def _justifyContent(inputString, justify, numCharacterCodes, color):
        if justify == "left":
            inputString = inputString.ljust(22 + numCharacterCodes, "^")
        elif justify == "right":
            inputString = inputString.rjust(22 + numCharacterCodes, "^")
        elif justify == "center":
            inputString = inputString.center(22 + numCharacterCodes, "^")
        inputString = inputString.replace("^", color)

        return inputString

    @staticmethod
    def _reverse_convert(charArray, normalize: bool = False):
        convertedArray = []
        for line in charArray:
            newLine = []
            for character in line:
                newLine.append(reverseCharacters[character])
            if normalize:
                convertedArray.append("".join(newLine))
            else:
                convertedArray.append(newLine)
        if normalize:
            return "\n".join(convertedArray)
        else:
            return convertedArray

    def _add_vestaboard_spacing(self, lines, size=None, justify="left"):
        if size is None:
            size = [6, 22]
        _, maxChars = size
        longestLine = 0
        for line in lines:
            lineLength = len(line)
            if lineLength > longestLine:
                longestLine = lineLength

        numSpacesNeeded = 0
        if maxChars > longestLine:
            numSpacesNeeded = math.floor((maxChars - longestLine) / 2)

        paddedLines = []
        for line in lines:
            lineLength = len(line)
            if justify == "left":
                paddedLines.append(
                    self.convertLine(
                        " " * numSpacesNeeded
                        + line
                        + " " * (maxChars - lineLength - numSpacesNeeded)
                    )
                )
            elif justify == "right":
                paddedLines.append(
                    self.convertLine(
                        " " * (maxChars - lineLength - numSpacesNeeded)
                        + line
                        + " " * numSpacesNeeded
                    )
                )

        return paddedLines

    def convertPlainText(
        self,
        text,
        size=None,
        justify="center",
        align="center",
        useVestaboardCentering=False,
    ):
        if size is None:
            size = [6, 22]
        maxRows, maxCols = size
        splitLines = []
        for linebreak in text.split("\n"):
            splitLines += textwrap.fill(linebreak, maxCols).split("\n")
        if len(splitLines) > maxRows:
            warningMessage = f"This text is too long to fit within the space specified ({maxRows} rows of {maxCols} characters each). Only the first {maxRows} rows are being returned."
            warnings.warn(warningMessage)

        convertedLines = []

        if useVestaboardCentering:
            if justify == "center":
                warnings.warn(
                    "Vestaboard formatting only affects left or right-justified text. Because of this, your text will be left justified by default. If you don't want your text left justified, remove `useVestaboardCentering` from your function call."
                )
                convertedLines = self._add_vestaboard_spacing(splitLines, size)
            else:
                convertedLines = self._add_vestaboard_spacing(splitLines, size, justify)
        else:
            for line in splitLines:
                convertedLines.append(self.convertLine(line, justify, "black", False))

        if len(convertedLines) < maxRows:
            blankLine = [0] * maxCols
            while len(convertedLines) < maxRows:
                if align == "bottom":
                    convertedLines.insert(0, blankLine)
                else:
                    convertedLines.append(blankLine)

                if len(convertedLines) < maxRows:
                    if align == "top":
                        convertedLines.append(blankLine)
                    else:
                        convertedLines.insert(0, blankLine)

        return convertedLines[0:maxRows]



import vestaboard.vbUrls as vbUrls

def test_post_url_matches():
    assert vbUrls.post == 'https://platform.vestaboard.com/subscriptions/{}/message'

def test_subscription_url_matches():
    assert vbUrls.subscription == 'https://platform.vestaboard.com/subscriptions'



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Vestaboard",
    version="1.2.4",
    author="Shane Sutro",
    author_email="shane@shanesutro.com",
    description="A Vestaboard Wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ShaneSutro/Vestaboard.git",
    packages=setuptools.find_packages(),
    license="MIT",
    install_requires=[
        'requests'
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
    ],
    python_requires='>=3.6',
)


# -*- coding: utf-8 -*-
"""Vestaboard Module

A lightweight wrapper for Vestaboard

Board - Class
Installable - Class
"""
import warnings
import json
import requests
from vestaboard.formatter import Formatter
import vestaboard.vbUrls as vbUrls


class Board:
    def __init__(
        self,
        installable=False,
        apiKey=False,
        apiSecret=False,
        subscriptionId=False,
        localApi: dict = None,
        readWrite: bool = False,
        validateKey: bool = False,
    ):
        """
        Returns an instance of Board().

        Args:
        installable - an instance of installable()
        apiKey - your Vestaboard API Key
        apiSecret - your Vestaboard API Secret
        subscriptionId - your Subscription ID (this can be obtained for you by creating a new installable() instance)
        readWrite - pass True to create a new instance of Read/Write Board
        validateKey - Read/Write & Local API only - sends a GET request to validate the API key and raises an error if the key is invalid

        Keyword Args:
        localApi = {
          'key': str - your local API key
          'ip': str - your boards local IP address
          'useSavedToken': bool - use saved token instead of provided key and IP
          'saveToken': bool - used alongside the enablement token to store your credentials. Defaults to True
          'enablementToken': str - pass your boards enablement token along with IP to
          enable your board's local API. Only pass in the enablement token if you are enabling a board
          or need a new local API key.
        }
        """
        if installable and not isinstance(installable, Installable):
            raise ValueError(
                f'Expected the first argument passed to be an instance of Installable, but instead got {type(installable)}. If you are not trying to pass an Installable, specify the name of the arguments, or pass "None" as the first argument.'
            )

        self.localOptions = localApi
        self.localKey = ""
        self.localIP = ""
        self.readWrite = readWrite
        self.validateKey = validateKey

        if self.localOptions is not None:
            if self.checkAndEnableLocalAPI():
                return
            if len(self.localOptions) < 2:
                if (
                    "useSavedToken" in self.localOptions
                    and self.localOptions["useSavedToken"]
                ):
                    token = get_local_token()
                    self.localKey = token[0]
                    self.localIP = token[1]
                else:
                    raise ValueError(
                        "Either the local API key or your board's IP address was not provided. If you have saved credentials, pass `useSavedToken` to use those credentials instead."
                    )
            else:
                if (
                    "useSavedToken" in self.localOptions
                    and self.localOptions["useSavedToken"]
                ):
                    token = get_local_token()
                    self.localKey = token[0]
                    self.localIP = token[1]
                    if "ip" in self.localOptions and self.localOptions["ip"]:
                        self.localIP = self.localOptions["ip"]
                else:
                    self.localKey = localApi["key"]
                    self.localIP = localApi["ip"]
            if self.validateKey:
                self.read()
            return

        if not installable:  # check for cred file
            if readWrite:
                if not apiKey:
                    try:
                        creds = get_creds()
                        if len(creds) < 2:
                            raise ValueError(
                                "You must supply a read/write enabled API key to use readWrite mode, or instantiate an Installable in readWrite mode to store your credentials."
                            )
                        self.apiKey = creds[0]
                        if self.validateKey:
                            self.read()
                    except FileNotFoundError as exc:
                        raise FileNotFoundError(
                            "I couldn't find any saved credentials, and no API key was provided.\nTo use readWrite mode, you must supply a read/write enabled API key, or instantiate an Installable in readWrite mode."
                        ) from exc
                else:
                    self.apiKey = apiKey
                    if self.validateKey:
                        self.read()
                    return
            if not apiKey or not apiSecret or not subscriptionId:
                try:
                    creds = get_creds()
                    if len(creds) == 3:
                        self.apiKey = creds[0]
                        self.apiSecret = creds[1]
                        self.subscriptionId = creds[2]
                    elif len(creds) < 3 or len(creds) > 3:
                        raise ValueError(
                            "Credentials have been saved, but one or more are missing. Create a new installable and pass in saveCredentials=True, or pass in all three parameters when initializing a new Board."
                        )
                except FileNotFoundError as exc:
                    raise ValueError(
                        "You must create an installable first or save credentials by passing saveCredentials=True into installable()."
                    ) from exc
            else:
                self.apiKey = apiKey
                self.apiSecret = apiSecret
                self.subscriptionId = subscriptionId
        else:
            self.apiKey = installable.apiKey
            self.apiSecret = installable.apiSecret
            self.subscriptionId = installable.subscriptionId or subscriptionId

    def post(self, text):
        if self.localOptions:
            self._post_local(text)
            return

        if self.readWrite:
            if not isinstance(text, str):
                raise ValueError(
                    "The `post()` method expects a plain string - if your characters are already converted, use the `raw()` method instead."
                )
            headers = {"X-Vestaboard-Read-Write-Key": self.apiKey}
            fm = Formatter()
            self.raw(
                fm.convertPlainText(text, justify="left", useVestaboardCentering=True)
            )
            return

        headers = {
            "X-Vestaboard-Api-Key": self.apiKey,
            "X-Vestaboard-Api-Secret": self.apiSecret,
        }
        finalText = Formatter()._standard(text)
        req = requests.post(
            vbUrls.post.format(self.subscriptionId),
            headers=headers,
            json=finalText,
            timeout=5,
        )
        req.raise_for_status()

    def raw(self, charList: list, pad=None):
        """
        Posts already-formatted characters to the board.

        Keyword arguments:

        `charList` - a list of character lists

        `pad` - adds padding when list or character lists is less than 6. Valid options are `top`, `bottom`, and `center`.
        """
        base_filler = [0] * 22
        filler_needed = 6 - len(charList)
        for i, row in enumerate(charList):
            if not isinstance(row, list):
                raise ValueError(f"Nested items must be lists, not {type(row)}.")
            if len(row) != 22:
                raise ValueError(
                    f"Nested lists must be exactly 22 characters long. Element at {i} is {len(row)} characters long. Use the Formatter().convertLine() function if you need to add padding to your row."
                )
            for j, char in enumerate(row):
                if not isinstance(char, int):
                    raise ValueError(
                        f"Nested lists must contain numbers only - check row {i} char {j} (0 indexed)"
                    )
        if len(charList) > 6:
            # warnings.warn doesn't work with f strings
            warning_message = f"The Vestaboard API accepts only 6 lines of characters; you've passed in {len(charList)}. Only the first 6 will be shown."
            warnings.warn(warning_message)
            del charList[6:]
        elif len(charList) < 6:
            if pad == "below":
                for i in range(filler_needed):
                    charList.append(base_filler)
            elif pad == "above":
                for i in range(filler_needed):
                    charList.insert(0, base_filler)
            else:
                if pad == None:
                    # warnings.warn doesn't work with f strings
                    warning_message = f'you provided a list with length {len(charList)}, which has been centered vertically on the board by default. Either provide a list with length 6, or set the "pad" option to suppress this warning.'
                    warnings.warn(warning_message)
                while len(charList) < 6:
                    charList.append(base_filler)
                    if len(charList) < 6:
                        charList.insert(0, base_filler)
        finalText = Formatter()._raw(charList)
        if self.localKey and self.localIP:
            self._raw_local(finalText["characters"])
        elif self.readWrite and self.apiKey:
            headers = {"X-Vestaboard-Read-Write-Key": self.apiKey}
            requests.post(
                vbUrls.readWrite,
                headers=headers,
                data=json.dumps(finalText["characters"]),
                timeout=5,
            )
        else:
            headers = {
                "X-Vestaboard-Api-Key": self.apiKey,
                "X-Vestaboard-Api-Secret": self.apiSecret,
            }
            requests.post(
                vbUrls.post.format(self.subscriptionId),
                headers=headers,
                json=finalText,
                timeout=5,
            )

    def _enableLocalApi(
        self, enablementKey: str = "", boardIP: str = "", saveToFile: bool = True
    ):
        headers = {"X-Vestaboard-Local-Api-Enablement-Token": enablementKey}
        try:
            response = requests.post(
                vbUrls.enableLocal.format(boardIP), headers=headers, timeout=5
            )
            if not response.ok:
                print(
                    "Looks like that didn't work. The Vestaboard returned the following message:\n",
                    response.text,
                )
            else:
                parsed = response.json()
                self.localOptions = {"useSavedToken": True}
                self.localIP = boardIP
                self.localKey = parsed["apiKey"]
                print("Success! Here's your local API token:\n", parsed["apiKey"])
                if saveToFile:
                    with open("./local.txt", "w", encoding="UTF-8") as local:
                        local.write(parsed["apiKey"])
                        local.write("\n")
                        local.write(boardIP)
                        print(
                            "Saved to ./local.txt! This instance of Board can now be used with the local API, or pass the `localApi={'useSavedToken': True}` when instantiating a new Board to use your saved credentials."
                        )
                        local.close()
        except OSError as e:
            raise ConnectionError(
                "I couldn't connect to your board. Are you on the same network as the board you're trying to connect to?",
                e,
            ) from e

    def checkAndEnableLocalAPI(self):
        if "enablementToken" not in self.localOptions:
            return False
        elif "ip" not in self.localOptions:
            raise ValueError(
                "Looks like you're trying to enable your board's local API, but your board's IP was not passed in.\nSupply both the enablement token from Vestaboard and your board's local IP address and try again with just the key and IP."
            )

        if "key" in self.localOptions:
            warnings.warn(
                "An enablement token was provided, so I'm enabling the local API for you. If you've already enabled the local API on your board, remove the enablement token and try again."
            )

        if "saveToken" in self.localOptions and self.localOptions["saveToken"] == False:
            self._enableLocalApi(
                self.localOptions["enablementToken"], self.localOptions["ip"], False
            )
            return True
        self._enableLocalApi(
            self.localOptions["enablementToken"], self.localOptions["ip"]
        )
        return True

    def read(self, options: dict = None):
        if options is None:
            options = {}
        if not self.readWrite:
            if self.localOptions is None or (
                "useSavedToken" not in self.localOptions
                and "key" not in self.localOptions
            ):
                raise ValueError(
                    '.read() is only available when using local API or by using a read/write enabled API key.\nPass "readWrite=True" along with your apiKey to enable readWrite mode.'
                )
        if self.localIP and self.localKey:
            localHeader = {"X-Vestaboard-Local-Api-Key": self.localKey}
            res = requests.get(
                vbUrls.postLocal.format(self.localIP), headers=localHeader, timeout=5
            )
            res.raise_for_status()
            if res.text == "" and not self.validateKey:
                warnings.warn(
                    "The request succeeded, but I didn't get any text back from your board. If your board is currently displaying a message and this is unexpected, ensure you are using a valid local API key rather than a regular API key from Vestaboard's website."
                )
                return
            elif res.text == "" and self.validateKey:
                warnings.warn(
                    "This API key appears to be invalid; I recieved an empty string from your board. If your board is not currently blank, double check your API key, or pass in your enablement token again to get a new API key."
                )
                return
            response_text = res.json()["message"]
        elif self.readWrite and self.apiKey:
            readWriteHeader = {"X-Vestaboard-Read-Write-Key": self.apiKey}
            print("Getting board message")
            res = requests.get(vbUrls.readWrite, headers=readWriteHeader, timeout=5)
            res.raise_for_status()
            response_text = json.loads(res.json()["currentMessage"]["layout"])
        if "print" in options and options["print"]:
            print(res.json())
        if "convert" in options and options["convert"]:
            if "normalize" in options and options["normalize"]:
                converted = Formatter()._reverse_convert(response_text, normalize=True)
                return converted
            else:
                converted = Formatter()._reverse_convert(response_text)
                return converted
        if self.validateKey:
            print("Your API key appears to be valid!")
        return res.json()

    def _post_local(self, text):
        localHeader = {"X-Vestaboard-Local-Api-Key": self.localKey}
        fm = Formatter()
        convertedByDefault = fm.convertPlainText(text)
        res = requests.post(
            vbUrls.postLocal.format(self.localIP),
            headers=localHeader,
            data=json.dumps(convertedByDefault),
            timeout=5,
        )
        res.raise_for_status()
        print(res.text)

    def _raw_local(self, chars):
        localHeader = {"X-Vestaboard-Local-Api-Key": self.localKey}
        res = requests.post(
            vbUrls.postLocal.format(self.localIP),
            headers=localHeader,
            data=json.dumps(chars),
            timeout=5,
        )
        res.raise_for_status()


class Installable:
    def __init__(
        self, apiKey=False, apiSecret=False, getSubscription=True, saveCredentials=True
    ):
        """
        Returns an instance of Installable()

        You can pass this into an instance of Board() as the first keyword argument.
        Keyword arguments:
        apiKey: String (required) - your Vestaboard API Key
        apiSecret: String (required) - your Vestaboard API Secret
        getSubscripion: Bool (optional, default True) - If you already have your subscription ID, you may pass False into this method

        saveCredentials: Bool (options, default True) - Choose whether or not to store your API keys in the home directory
        """
        self.apiKey = apiKey
        self.apiSecret = apiSecret
        self.saveCredentials = saveCredentials
        if not apiKey or not apiSecret:
            raise ValueError(
                "Installables must have an apiKey and apiSecret parameter."
            )
        if saveCredentials and apiKey and apiSecret:
            with open("./credentials.txt", "w", encoding="UTF-8") as cred:
                cred.write(apiKey + "\n")
                cred.write(apiSecret + "\n")
                cred.close()
        if getSubscription:
            self.subscriptionId = self.get_subscriptions(saveCredentials)[0]["_id"]
        else:
            self.subscriptionId = False

    def get_subscriptions(self, save=True):
        if not self.apiKey or not self.apiSecret:
            raise ValueError(
                "There are no keys set. Call Installable(apiKey= apiSecret=) with your installable keys."
            )
        headers = {
            "X-Vestaboard-Api-Key": self.apiKey,
            "X-Vestaboard-Api-Secret": self.apiSecret,
        }
        response = requests.get(vbUrls.subscription, headers=headers, timeout=5)
        response.raise_for_status()
        if self.saveCredentials or save and response.status_code == 200:
            with open("./credentials.txt", "a", encoding="UTF-8") as cred:
                cred.write(response.json()["subscriptions"][0]["_id"] + "\n")
                cred.close()

        print("Subscription id(s):", response.json()["subscriptions"])
        return response.json()["subscriptions"]


def get_creds():
    with open("./credentials.txt", "r", encoding="UTF-8") as cred:
        creds = cred.read().splitlines()
        return creds


def get_local_token():
    try:
        with open("./local.txt", "r", encoding="UTF-8") as local:
            token = local.read().splitlines()
            if len(token) == 0:
                raise ValueError(
                    "Either your local token or IP was not passed in, and there's nothing stored.\nEither pass in a value for `localKey` and `localIP`, or store a copy of your local key and IP with the name 'local.txt.' Don't forget to include your board's IP address on line 2."
                )
            elif len(token) == 1:
                raise ValueError(
                    "Your local.txt file is missing either the local key or your board's IP address. Ensure your key is on line 1 and your IP is on line 2 and try again."
                )
            elif len(token) > 2:
                raise ValueError(
                    "Your local.txt file contains more than your key and IP (or has more than two lines of text). Please remove the extra key or line and try again. Your local.txt file should contain only the local key and board IP address, each on their own line."
                )
            return (token[0], token[1])
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            "I couldn't find any stored tokens. If you have already enabled your board's local API, you can pass in your board's IP and token in via the `localApi` dict, or pass in just the enablement token to get and store a new API token."
        ) from exc
